@@ -44,13 +44,14 @@ use mimalloc::MiMalloc;
 static GLOBAL: MiMalloc = MiMalloc;
 
 use atty::Stream;
-use clap::{crate_version, App, AppSettings, Arg};
+use clap::{crate_authors, crate_description, crate_version, App, AppSettings, Arg};
 use sliders::fix_all_sliders;
 use std::{env, path::Path};
+use style::BackgroundColor;
 use summary::DiffResult;
 use syntax::init_next_prev;
 use typed_arena::Arena;
-use unchanged::skip_unchanged_at_ends;
+use unchanged::skip_unchanged;
 use walkdir::WalkDir;
 
 use crate::{
@@ -64,18 +65,39 @@ use crate::{
 
 extern crate pretty_env_logger;
 
-fn configure_color() {
-    if atty::is(Stream::Stdout) || env::var("GIT_PAGER_IN_USE").is_ok() {
-        // Always enable colour if stdout is a TTY or if the git pager is active.
-        // TODO: provide a way to disable this.
-        // TODO: consider following the env parsing logic in git_config_bool
-        // in config.c.
-        colored::control::set_override(true);
-    }
+/// Choose the display width: try to autodetect, or fall back to a
+/// sensible default.
+fn detect_display_width() -> usize {
+    term_size::dimensions().map(|(w, _)| w).unwrap_or(80)
+}
+
+fn configure_color(color_output: ColorOutput) {
+    let enable_color = match color_output {
+        ColorOutput::Always => true,
+        ColorOutput::Auto => {
+            // Always enable colour if stdout is a TTY or if the git pager is active.
+            // TODO: consider following the env parsing logic in git_config_bool
+            // in config.c.
+            atty::is(Stream::Stdout) || env::var("GIT_PAGER_IN_USE").is_ok()
+        }
+        ColorOutput::Never => false,
+    };
+
+    colored::control::set_override(enable_color);
+}
+
+enum ColorOutput {
+    Always,
+    Auto,
+    Never,
 }
 
 enum Mode {
     Diff {
+        print_unchanged: bool,
+        background_color: BackgroundColor,
+        color_output: ColorOutput,
+        display_width: usize,
         display_path: String,
         lhs_path: String,
         rhs_path: String,
@@ -88,54 +110,94 @@ enum Mode {
     },
 }
 
+fn app() -> clap::App<'static> {
+    App::new("Difftastic")
+        .version(crate_version!())
+        .about(crate_description!())
+        .author(crate_authors!())
+        .after_long_help(concat!(
+            "You can compare two files with difftastic by specifying them as arguments.\n\n",
+            "$ ",
+            env!("CARGO_BIN_NAME"),
+            " old.js new.js\n\n",
+            "You can also use directories as arguments. Difftastic will walk both directories and compare files with matching names.\n\n",
+            "$ ",
+            env!("CARGO_BIN_NAME"),
+            " old/ new/\n\n",
+            "Difftastic can also be invoked with 7 arguments in the format that GIT_EXTERNAL_DIFF expects.\n\n",
+            "See the full manual at: https://difftastic.wilfred.me.uk/")
+        )
+        .arg(
+            Arg::new("dump-syntax")
+                .long("dump-syntax")
+                .takes_value(true)
+                .value_name("PATH")
+                .long_help(
+                    "[DEBUG] Parse a single file with tree-sitter and display the difftastic syntax tree.",
+                ),
+        )
+        .arg(
+            Arg::new("dump-ts")
+                .long("dump-ts")
+                .takes_value(true)
+                .value_name("PATH")
+                .long_help(
+                    "[DEBUG] Parse a single file with tree-sitter and display the tree-sitter parse tree.",
+                ),
+        )
+        .arg(
+            Arg::new("width")
+                .long("width")
+                .takes_value(true)
+                .value_name("COLUMNS")
+                .long_help("Use this many columns when calculating line wrapping. Overrides $DFT_WIDTH if present. If not specified, difftastic will detect the terminal width.")
+                .validator(|s| s.parse::<usize>())
+                .required(false),
+        )
+        .arg(
+            Arg::new("color").long("color")
+                .possible_values(["always", "auto", "never"])
+                .value_name("WHEN")
+                .help("When to use color output.")
+        )
+        .arg(
+            Arg::new("background").long("background")
+                .value_name("BACKGROUND")
+                .possible_values(["dark", "light"])
+                .help("Set the background color. Overrides $DFT_BACKGROUND if present. Difftastic will prefer brighter colours on dark backgrounds.")
+        )
+        .arg(
+            Arg::new("skip-unchanged").long("skip-unchanged")
+                .help("Don't display anything if a file is unchanged.")
+        )
+        .arg(
+            Arg::new("paths")
+                .value_name("PATHS")
+                .multiple_values(true)
+                .allow_invalid_utf8(true),
+        )
+        .setting(AppSettings::ArgRequiredElseHelp)
+}
+
 /// Parse CLI arguments passed to the binary.
 fn parse_args() -> Mode {
-    let matches =
-        App::new("Difftastic")
-            .version(crate_version!())
-            .about("A syntax aware diff.")
-            .author("Wilfred Hughes")
-            .arg(Arg::with_name("dump-syntax").long("dump-syntax").help(
-                "Parse a single file with tree-sitter and display the difftastic syntax tree.",
-            ))
-            .arg(Arg::with_name("dump-ts").long("dump-ts").help(
-                "Parse a single file with tree-sitter and display the tree-sitter parse tree.",
-            ))
-            .arg(Arg::with_name("positional_args").multiple(true))
-            .setting(AppSettings::ArgRequiredElseHelp)
-            .get_matches();
+    let matches = app().get_matches();
 
-    let args: Vec<_> = matches.values_of_lossy("positional_args").unwrap();
+    if let Some(path) = matches.value_of("dump-syntax") {
+        return Mode::DumpSyntax {
+            path: path.to_string(),
+        };
+    }
+
+    if let Some(path) = matches.value_of("dump-ts") {
+        return Mode::DumpTreeSitter {
+            path: path.to_string(),
+        };
+    }
+
+    let args: Vec<_> = matches.values_of_lossy("paths").unwrap();
     info!("CLI arguments: {:?}", args);
 
-    if matches.is_present("dump-syntax") {
-        if args.len() == 1 {
-            return Mode::DumpSyntax {
-                path: args[0].clone(),
-            };
-        } else {
-            // TODO: delegate this parsing to clap.
-            panic!(
-                "Error: --dump-syntax takes one argument, but got: {}",
-                args.len()
-            );
-        }
-    }
-
-    if matches.is_present("dump-ts") {
-        if args.len() == 1 {
-            return Mode::DumpTreeSitter {
-                path: args[0].clone(),
-            };
-        } else {
-            // TODO: delegate this parsing to clap.
-            panic!(
-                "Error: --dump-ts takes one argument, but got: {}",
-                args.len()
-            );
-        }
-    }
-
     // TODO: document these different ways of calling difftastic.
     let (display_path, lhs_path, rhs_path) = match &args[..] {
         [lhs_path, rhs_path] => (
@@ -162,14 +224,62 @@ fn parse_args() -> Mode {
                 rhs_tmp_file.to_string(),
             )
         }
-        _ => panic!(
-            "Unexpected number of arguments, got {}: {:?}",
-            args.len(),
-            args
-        ),
+        _ => {
+            eprintln!("error: Difftastic does not support being called with {} arguments. See --help for more information.", args.len());
+            std::process::exit(1);
+        }
     };
 
+    let display_width = if let Some(arg_width) = matches.value_of("width") {
+        arg_width
+            .parse::<usize>()
+            .expect("Already validated by clap")
+    } else {
+        let env_width = if let Ok(env_width) = env::var("DFT_WIDTH") {
+            env_width.parse::<usize>().ok()
+        } else {
+            None
+        };
+        env_width.unwrap_or_else(detect_display_width)
+    };
+
+    let color_output = if let Some(color_when) = matches.value_of("color") {
+        if color_when == "always" {
+            ColorOutput::Always
+        } else if color_when == "never" {
+            ColorOutput::Never
+        } else {
+            ColorOutput::Auto
+        }
+    } else {
+        ColorOutput::Auto
+    };
+
+    let background_color = if let Some(background) = matches.value_of("background") {
+        if background == "light" {
+            BackgroundColor::Light
+        } else {
+            BackgroundColor::Dark
+        }
+    } else {
+        if let Ok(background) = env::var("DFT_BACKGROUND") {
+            if background == "light" {
+                BackgroundColor::Light
+            } else {
+                BackgroundColor::Dark
+            }
+        } else {
+            BackgroundColor::Dark
+        }
+    };
+
+    let print_unchanged = !matches.is_present("skip-unchanged");
+
     Mode::Diff {
+        print_unchanged,
+        background_color,
+        color_output,
+        display_width,
         display_path,
         lhs_path,
         rhs_path,
@@ -192,9 +302,7 @@ fn reset_sigpipe() {
 /// The entrypoint.
 fn main() {
     pretty_env_logger::init();
-
     reset_sigpipe();
-    configure_color();
 
     match parse_args() {
         Mode::DumpTreeSitter { path } => {
@@ -231,20 +339,37 @@ fn main() {
             }
         }
         Mode::Diff {
+            print_unchanged,
+            background_color,
+            color_output,
+            display_width,
             display_path,
             lhs_path,
             rhs_path,
+            ..
         } => {
+            configure_color(color_output);
+
             let lhs_path = Path::new(&lhs_path);
             let rhs_path = Path::new(&rhs_path);
 
             if lhs_path.is_dir() && rhs_path.is_dir() {
                 for diff_result in diff_directories(lhs_path, rhs_path) {
-                    print_diff_result(&diff_result);
+                    print_diff_result(
+                        display_width,
+                        background_color,
+                        print_unchanged,
+                        &diff_result,
+                    );
                 }
             } else {
                 let diff_result = diff_file(&display_path, lhs_path, rhs_path);
-                print_diff_result(&diff_result);
+                print_diff_result(
+                    display_width,
+                    background_color,
+                    print_unchanged,
+                    &diff_result,
+                );
             }
         }
     };
@@ -280,8 +405,9 @@ fn diff_file_content(display_path: &str, lhs_bytes: &[u8], rhs_bytes: &[u8]) ->
     // TODO: take a Path directly instead.
     let path = Path::new(&display_path);
 
-    // Take the longer of the two files when guessing language. This
-    // is useful when we've added or removed a whole file.
+    // Take the larger of the two files when guessing the
+    // language. This is useful when we've added or removed a whole
+    // file.
     let guess_src = if lhs_src.len() > rhs_src.len() {
         &lhs_src
     } else {
@@ -309,7 +435,7 @@ fn diff_file_content(display_path: &str, lhs_bytes: &[u8], rhs_bytes: &[u8]) ->
 
             init_all_info(&lhs, &rhs);
 
-            let (possibly_changed_lhs, possibly_changed_rhs) = skip_unchanged_at_ends(&lhs, &rhs);
+            let (possibly_changed_lhs, possibly_changed_rhs) = skip_unchanged(&lhs, &rhs);
             init_next_prev(&possibly_changed_lhs);
             init_next_prev(&possibly_changed_rhs);
 
@@ -366,9 +492,19 @@ fn diff_directories(lhs_dir: &Path, rhs_dir: &Path) -> Vec<DiffResult> {
     res
 }
 
-fn print_diff_result(summary: &DiffResult) {
+fn print_diff_result(
+    display_width: usize,
+    background: BackgroundColor,
+    print_unchanged: bool,
+    summary: &DiffResult,
+) {
     if summary.binary {
-        println!("{}", style::header(&summary.path, 1, 1, "binary"));
+        if print_unchanged {
+            println!(
+                "{}",
+                style::header(&summary.path, 1, 1, "binary", background)
+            );
+        }
         return;
     }
 
@@ -386,18 +522,21 @@ fn print_diff_result(summary: &DiffResult) {
 
     let lang_name = summary.language.clone().unwrap_or_else(|| "text".into());
     if hunks.is_empty() {
-        println!("{}", style::header(&summary.path, 1, 1, &lang_name));
-        if lang_name == "text" {
-            println!("No changes.\n");
-        } else {
-            println!("No syntactic changes.\n");
+        if print_unchanged {
+            println!(
+                "{}",
+                style::header(&summary.path, 1, 1, &lang_name, background)
+            );
+            if lang_name == "text" {
+                println!("No changes.\n");
+            } else {
+                println!("No syntactic changes.\n");
+            }
         }
         return;
     }
 
     if env::var("INLINE").is_ok() {
-        println!("{}", style::header(&summary.path, 1, 1, &lang_name));
-
         println!(
             "{}",
             inline::display(
@@ -405,7 +544,10 @@ fn print_diff_result(summary: &DiffResult) {
                 &summary.rhs_src,
                 &summary.lhs_positions,
                 &summary.rhs_positions,
-                &hunks
+                &hunks,
+                &summary.path,
+                &lang_name,
+                background
             )
         );
     } else {
@@ -413,6 +555,8 @@ fn print_diff_result(summary: &DiffResult) {
             "{}",
             side_by_side::display_hunks(
                 &hunks,
+                display_width,
+                background,
                 &summary.path,
                 &lang_name,
                 &summary.lhs_src,
@@ -437,4 +581,15 @@ mod tests {
         assert_eq!(res.rhs_positions, vec![]);
         assert!(!res.binary);
     }
+
+    #[test]
+    fn test_app() {
+        app().debug_assert();
+    }
+
+    #[test]
+    fn test_detect_display_width() {
+        // Basic smoke test.
+        assert!(detect_display_width() > 10);
+    }
 }
