@@ -4,33 +4,19 @@ use colored::{Color, Colorize};
 use std::{
     cmp::max,
     collections::{HashMap, HashSet},
-    env,
 };
 
 use crate::{
-    context::opposite_positions,
-    hunks::{aligned_lines_from_hunk, Hunk},
-    lines::{codepoint_len, format_line_num, LineNumber, MaxLine},
+    context::all_matched_lines_filled,
+    hunks::{matched_lines_for_hunk, Hunk},
+    lines::{codepoint_len, format_line_num, LineNumber},
     positions::SingleLineSpan,
-    style::{self, apply_colors, color_positions, split_and_apply, Style},
+    style::{self, apply_colors, color_positions, split_and_apply, BackgroundColor, Style},
     syntax::{zip_pad_shorter, MatchedPos},
 };
 
 const SPACER: &str = " ";
 
-/// Choose the display width: honour environment variables, then try
-/// to autodetect, or fall back to a sensible default.
-fn display_width() -> usize {
-    // TODO: document in manual.
-    if let Ok(s) = env::var("DFT_WIDTH") {
-        if let Ok(i) = s.parse::<usize>() {
-            return i;
-        }
-    }
-
-    term_size::dimensions().map(|(w, _)| w).unwrap_or(80)
-}
-
 /// Split `s` by newlines. Always returns a non-empty vec.
 ///
 /// This differs from `str::lines`, which considers `""` to be zero
@@ -54,16 +40,31 @@ fn format_missing_line_num(prev_num: LineNumber, column_width: usize) -> String
         ".".repeat(num_digits),
         width = column_width - 1
     )
+    .dimmed()
+    .to_string()
 }
 
 /// Display `src` in a single column (e.g. a file removal or addition).
-fn display_single_column(display_path: &str, lang_name: &str, src: &str, color: Color) -> String {
+fn display_single_column(
+    display_path: &str,
+    lang_name: &str,
+    src: &str,
+    is_lhs: bool,
+    background: BackgroundColor,
+) -> String {
     let column_width = format_line_num(src.lines().count().into()).len();
 
     let mut result = String::with_capacity(src.len());
-    result.push_str(&style::header(display_path, 1, 1, lang_name));
+    result.push_str(&style::header(display_path, 1, 1, lang_name, background));
     result.push('\n');
 
+    let color = match (is_lhs, background) {
+        (true, BackgroundColor::Dark) => Color::BrightRed,
+        (true, BackgroundColor::Light) => Color::Red,
+        (false, BackgroundColor::Dark) => Color::BrightGreen,
+        (false, BackgroundColor::Light) => Color::Green,
+    };
+
     for (i, line) in src.lines().enumerate() {
         result.push_str(
             &format_line_num_padded(i.into(), column_width)
@@ -81,6 +82,7 @@ fn display_line_nums(
     lhs_line_num: Option<LineNumber>,
     rhs_line_num: Option<LineNumber>,
     widths: &Widths,
+    background: BackgroundColor,
     lhs_has_novel: bool,
     rhs_has_novel: bool,
     prev_lhs_line_num: Option<LineNumber>,
@@ -90,7 +92,12 @@ fn display_line_nums(
         Some(line_num) => {
             let s = format_line_num_padded(line_num, widths.lhs_line_nums);
             if lhs_has_novel {
-                s.bright_red().to_string()
+                // TODO: factor out applying colours to line numbers.
+                match background {
+                    BackgroundColor::Dark => s.bright_red(),
+                    BackgroundColor::Light => s.red(),
+                }
+                .to_string()
             } else {
                 s
             }
@@ -104,7 +111,11 @@ fn display_line_nums(
         Some(line_num) => {
             let s = format_line_num_padded(line_num, widths.rhs_line_nums);
             if rhs_has_novel {
-                s.bright_green().to_string()
+                match background {
+                    BackgroundColor::Dark => s.bright_green(),
+                    BackgroundColor::Light => s.green(),
+                }
+                .to_string()
             } else {
                 s
             }
@@ -166,7 +177,7 @@ impl Widths {
     }
 }
 
-fn lines_with_novel(
+pub fn lines_with_novel(
     lhs_mps: &[MatchedPos],
     rhs_mps: &[MatchedPos],
 ) -> (HashSet<LineNumber>, HashSet<LineNumber>) {
@@ -187,13 +198,14 @@ fn lines_with_novel(
 /// Calculate positions of highlights on both sides. This includes
 /// both syntax highlighting and added/removed content highlighting.
 fn highlight_positions(
+    background: BackgroundColor,
     lhs_mps: &[MatchedPos],
     rhs_mps: &[MatchedPos],
 ) -> (
     HashMap<LineNumber, Vec<(SingleLineSpan, Style)>>,
     HashMap<LineNumber, Vec<(SingleLineSpan, Style)>>,
 ) {
-    let lhs_positions = color_positions(true, lhs_mps);
+    let lhs_positions = color_positions(true, background, lhs_mps);
     // Preallocate the hashmap assuming the average line will have 2 items on it.
     let mut lhs_styles: HashMap<LineNumber, Vec<(SingleLineSpan, Style)>> =
         HashMap::with_capacity(lhs_positions.len() / 2);
@@ -202,7 +214,7 @@ fn highlight_positions(
         styles.push((span, style));
     }
 
-    let rhs_positions = color_positions(false, rhs_mps);
+    let rhs_positions = color_positions(false, background, rhs_mps);
     let mut rhs_styles: HashMap<LineNumber, Vec<(SingleLineSpan, Style)>> =
         HashMap::with_capacity(rhs_positions.len() / 2);
     for (span, style) in rhs_positions {
@@ -239,6 +251,8 @@ fn highlight_as_novel(
 
 pub fn display_hunks(
     hunks: &[Hunk],
+    display_width: usize,
+    background: BackgroundColor,
     display_path: &str,
     lang_name: &str,
     lhs_src: &str,
@@ -246,26 +260,18 @@ pub fn display_hunks(
     lhs_mps: &[MatchedPos],
     rhs_mps: &[MatchedPos],
 ) -> String {
-    let opposite_to_lhs = opposite_positions(lhs_mps);
-    let opposite_to_rhs = opposite_positions(rhs_mps);
-
-    let lhs_colored_src = apply_colors(lhs_src, true, lhs_mps);
-    let rhs_colored_src = apply_colors(rhs_src, false, rhs_mps);
+    let lhs_colored_src = apply_colors(lhs_src, true, background, lhs_mps);
+    let rhs_colored_src = apply_colors(rhs_src, false, background, rhs_mps);
 
     if lhs_src.is_empty() {
-        return display_single_column(
-            display_path,
-            lang_name,
-            &rhs_colored_src,
-            Color::BrightGreen,
-        );
+        return display_single_column(display_path, lang_name, &rhs_colored_src, false, background);
     }
     if rhs_src.is_empty() {
-        return display_single_column(display_path, lang_name, &lhs_colored_src, Color::BrightRed);
+        return display_single_column(display_path, lang_name, &lhs_colored_src, true, background);
     }
 
     // TODO: this is largely duplicating the `apply_colors` logic.
-    let (lhs_highlights, rhs_highlights) = highlight_positions(lhs_mps, rhs_mps);
+    let (lhs_highlights, rhs_highlights) = highlight_positions(background, lhs_mps, rhs_mps);
     let lhs_lines = split_on_newlines(lhs_src);
     let rhs_lines = split_on_newlines(rhs_src);
     let lhs_colored_lines = split_on_newlines(&lhs_colored_src);
@@ -276,23 +282,25 @@ pub fn display_hunks(
     let mut prev_lhs_line_num = None;
     let mut prev_rhs_line_num = None;
 
+    let matched_lines = all_matched_lines_filled(lhs_mps, rhs_mps);
+
     let mut out_lines: Vec<String> = vec![];
 
     for (i, hunk) in hunks.iter().enumerate() {
-        out_lines.push(style::header(display_path, i + 1, hunks.len(), lang_name));
+        out_lines.push(style::header(
+            display_path,
+            i + 1,
+            hunks.len(),
+            lang_name,
+            background,
+        ));
 
-        let aligned_lines = aligned_lines_from_hunk(
-            hunk,
-            &opposite_to_lhs,
-            &opposite_to_rhs,
-            lhs_src.max_line(),
-            rhs_src.max_line(),
-        );
-        let no_lhs_changes = hunk.lines.iter().all(|(l, _)| l.is_none());
-        let no_rhs_changes = hunk.lines.iter().all(|(_, r)| r.is_none());
+        let aligned_lines = matched_lines_for_hunk(&matched_lines, hunk);
+        let no_lhs_changes = hunk.novel_lhs.is_empty();
+        let no_rhs_changes = hunk.novel_rhs.is_empty();
         let same_lines = aligned_lines.iter().all(|(l, r)| l == r);
 
-        let widths = Widths::new(display_width(), &aligned_lines, &lhs_lines, &rhs_lines);
+        let widths = Widths::new(display_width, &aligned_lines, &lhs_lines, &rhs_lines);
         for (lhs_line_num, rhs_line_num) in aligned_lines {
             let lhs_line_novel = highlight_as_novel(
                 lhs_line_num,
@@ -311,6 +319,7 @@ pub fn display_hunks(
                 lhs_line_num,
                 rhs_line_num,
                 &widths,
+                background,
                 lhs_line_novel,
                 rhs_line_novel,
                 prev_lhs_line_num,
@@ -357,7 +366,7 @@ pub fn display_hunks(
             } else {
                 let lhs_line = match lhs_line_num {
                     Some(lhs_line_num) => split_and_apply(
-                        &lhs_lines[lhs_line_num.0],
+                        lhs_lines[lhs_line_num.0],
                         widths.lhs_content,
                         lhs_highlights.get(&lhs_line_num).unwrap_or(&vec![]),
                     ),
@@ -365,7 +374,7 @@ pub fn display_hunks(
                 };
                 let rhs_line = match rhs_line_num {
                     Some(rhs_line_num) => split_and_apply(
-                        &rhs_lines[rhs_line_num.0],
+                        rhs_lines[rhs_line_num.0],
                         widths.rhs_content,
                         rhs_highlights.get(&rhs_line_num).unwrap_or(&vec![]),
                     ),
@@ -388,7 +397,11 @@ pub fn display_hunks(
                         );
                         if let Some(line_num) = lhs_line_num {
                             if lhs_lines_with_novel.contains(&line_num) {
-                                s = s.bright_red().to_string()
+                                s = match background {
+                                    BackgroundColor::Dark => s.bright_red(),
+                                    BackgroundColor::Light => s.red(),
+                                }
+                                .to_string();
                             }
                         }
                         s
@@ -403,7 +416,11 @@ pub fn display_hunks(
                         );
                         if let Some(line_num) = rhs_line_num {
                             if rhs_lines_with_novel.contains(&line_num) {
-                                s = s.bright_green().to_string();
+                                s = match background {
+                                    BackgroundColor::Dark => s.bright_green(),
+                                    BackgroundColor::Light => s.green(),
+                                }
+                                .to_string();
                             }
                         }
                         s
@@ -450,16 +467,16 @@ mod tests {
         assert_eq!(widths.rhs_line_nums, 3);
     }
 
-    #[test]
-    fn test_display_width() {
-        // Basic smoke test.
-        assert!(display_width() > 10);
-    }
-
     #[test]
     fn test_display_single_column() {
         // Basic smoke test.
-        let res = display_single_column("foo.py", "Python", "print(123)\n", Color::Green);
+        let res = display_single_column(
+            "foo.py",
+            "Python",
+            "print(123)\n",
+            false,
+            BackgroundColor::Dark,
+        );
         assert!(res.len() > 10);
     }
 
@@ -515,12 +532,16 @@ mod tests {
         }];
 
         let hunks = [Hunk {
+            novel_lhs: HashSet::from([0.into()]),
+            novel_rhs: HashSet::from([0.into()]),
             lines: vec![(Some(0.into()), Some(0.into()))],
         }];
 
         // Simple smoke test.
         display_hunks(
             &hunks,
+            80,
+            BackgroundColor::Dark,
             "foo.el",
             "Emacs Lisp",
             "foo",
